/* som.c    main for som-language
	P. Chongstitvatana
	22 Oct 2002

	start Som project   			13 Sept 2003
	som v1  public release 			4 Dec 2004
	som-v15 rev2  seperate optim 	5 Mar 2005
	som-v16: hex, i-array, file i/o 6 Nov 2005
	som-v17 t-code					23 Dec 2005
	som-v17a						27 May 2006
	som-v19v						20 Apr 2007
	som-v5							7 Oct 2009
	som-v51							23 Dec 2010
*/

#include "compile.h"
#include <time.h>

FILE *FO, *FI;				// output/input file
int CP;						// current code pointer
int call_load;				// ads of "call loadfile"
int load_ads, cs_ads;		// ads of "loadfile", "CS" in compiler

extern int *CS;

void error(char *mess){
	printf("%s\n",mess);
	exit(0);
}

// make file.lst file.obj from file.xxx
PRIVATE void makename( char *src){
	int n;
	char obj[80], lst[80];
	n = strcspn(src,".");
	strncpy(obj,src,n);
	strcpy(obj+n,".obj");
	strncpy(lst,src,n);
	strcpy(lst+n,".lst");
	strpack(som_src,src);
	strpack(som_obj,obj);
	strpack(som_lst,lst);
}

PRIVATE int encode(int op, int c, int a){
	return (a << 16) | (c << 6) | op;
}

PRIVATE int readCS(FILE *fi){
	int i, start, end, op, c, a, b, k, n;
	fscanf(fi,"%d%d",&start,&end);
	for(i=start; i<=end; i+=2){
		fscanf(fi,"%d%d%d%d%d",&k,&op,&c,&a,&b);
		CS[i] = encode(op,c,a);
		CS[i+1] = b;
		if(op == tCase){
			fscanf(fi,"%d%d",&a,&n); 	// size of jtab
			CS[i+2] = n;
			for(k=0; k<n-3; k++){
				fscanf(fi,"%d%d",&a,&b);
				CS[i+3+k] = b;
			}
			i += n-2;
		}
	}
	return end;
}

PRIVATE int readData(FILE *fi){
	int i, start, end, c;
	fscanf(fi,"%d%d",&start,&end);
	for(i=start; i<=end; i++){
		fscanf(fi,"%d",&c);
		M[i] = c;
	}
	return end;
}

// read symbol to find binding of: load, cs
// read link vectors if exists, otherwise read from symtab
// link vector format is: 2 load_ads cs_ads
PRIVATE void readSym(FILE *fi){
	int i, n;
	char name[NAMELEN];
	int type,ref,arg,v;

	fscanf(fi,"%d",&n);
	if (n > 0)				// link vectors
		fscanf(fi,"%d%d",&load_ads,&cs_ads);
	else{					// read from symtab
		fscanf(fi,"%d",&n);
		for(i=0; i<n; i++){
			fscanf(fi,"%s%d%d%d%d",
				name,&type,&ref,&arg,&v);
			if(eqs(name,"loadfile"))
				load_ads = ref;
			else if(eqs(name,"CS"))
				cs_ads = ref;
		}
	}
//	printf("loadads %d  csads %d\n",load_ads,cs_ads);
}

PRIVATE void genCall(int ref){
	call_load = CP;
	CS[CP] = encode(tCall,0,1);	// fn in reg 1
	CS[CP+1] = ref;
	CS[CP+2] = encode(tSys,13,0);
	CS[CP+4] = 0;
	CP += 5;
}

// read object file, code and data
PRIVATE void readobj(FILE *fi){
	int magic, a;
	fscanf(fi,"%d",&magic);
	if(magic != OBJ_SOM51)
		error("not som 5.1 object");
	a = readCS(fi);
//	printf("load CS %d\n",a);
	CP = a+1;
	a = readData(fi);
//	printf("load DS %d\n",a);
	DS = a+1;			// free DS
}

extern int noi;

PRIVATE void execute(char *obj){
	int i, d;
    FI = fopen(obj,"r");
	readobj(FI);
	fclose(FI);
	d = DS;
	eval(2);
	DS = d;
}

PRIVATE void doSom(char *obj){
    FI = fopen(obj,"r");
	printf("load %s\n",obj);
	readobj(FI);
	readSym(FI);
	fclose(FI);
	genCall(load_ads);
	eval(2);
}

// ---------- string functions ----------

// pack C-string (char *) to S-string in M[] indexed by s1
void strpack(int s1, char *s2){
	uchar c1, c2, c3, c4;
	while( *s2 != 0 ){
		c1 = (*s2 != 0) ? *s2++ : 0;
		c2 = (*s2 != 0) ? *s2++ : 0;
		c3 = (*s2 != 0) ? *s2++ : 0;
		c4 = (*s2 != 0) ? *s2++ : 0;
		M[s1] = (((((((int)c1)<<8)|(int)c2)<<8)|
			(int)c3)<<8)|(int)c4;
		s1++;
	}
	M[s1] = 0;
}

// convert som-string to C-string
void strunpack(char *s1, int s2){
	int s;
	char c1, c2, c3, c4;
	while( M[s2] != 0 ){
		s = M[s2];
		c1 = (char)((s >> 24) & 255);
		c2 = (char)((s >> 16) & 255);
  		c3 = (char)((s >> 8) & 255);
   		c4 = (char)(s & 255);
   		if( c1 != 0 ) *s1++ = c1;
		if( c2 != 0 ) *s1++ = c2;
		if( c3 != 0 ) *s1++ = c3;
		if( c4 != 0 ) *s1++ = c4;
		s2++;
	}
	*s1 = 0;
}

void prstr(int i){
	int c1,c2,c3,c4;
	int k;				// limit at 100
	k = 0;
	while( M[i] != 0 && k < 100){
		c1 = (M[i] >> 24) & 255;
		c2 = (M[i] >> 16) & 255;
		c3 = (M[i] >> 8) & 255;
		c4 = M[i] & 255;
		if( c1 != 0 ) printf("%c",c1);
		if( c2 != 0 ) printf("%c",c2);
		if( c3 != 0 ) printf("%c",c3);
		if( c4 != 0 ) printf("%c",c4);
		i++; k++;
	}
}

// -------------------------

void Usage(void){
	printf("%s\n%s\n%s\n%s\n",
		"som51          interactive mode",
		"som51 file     compile and run, output .lst and .obj",
		"som51 -x file  execute obj file",
		"som51 -?       this help");
	exit(0);
}

// mode 1 interactive, 2 compile&run, 3 execute
int main( int argc, char *argv[] ){
	char source[32];
	int i, mode;

	mode = 2;
	for (i=1; i<argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			case '?': Usage(); break;
			case 'x': mode = 3; break;
//			case 'c': mode = 4; break; // test compiler object
			default:
				fprintf(stderr,"unknown option %s\n",argv[i]);
				exit(0);
			}
		}else
			break;   // first non option
	}
	if(argc == 1) mode = 1;
	initfile();
	switch(mode){
	case 1:
		M[MODE_ADS] = 1;
		doSom("som51-0.obj");
		break;
	case 2:
		M[MODE_ADS] = 2;
		makename(argv[i]);
		doSom("som51-0.obj");
		break;
	case 3:
		// if user program use "lex", it must set
		// M[lexmode_ads] 1 or 2, mostly 2 (batch)
		execute(argv[i]);
		break;
//	case 4:
//		M[MODE_ADS] = 2;
//		makename(argv[i]);
//		doSom("som51.obj");
//		break;
	}
	return 0;
}

//	time1 = (double)clock();
//	for(i=0; i<100; i++)
//	eval(1);
//	time2 = (double)clock();
//	printf("time %f\n",time2-time1);

